/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import static java.util.stream.Collectors.toMap;
import static javafx.collections.FXCollections.observableArrayList;
import static org.apache.commons.lang3.SystemUtils.IS_OS_LINUX;

import java.util.Map;
import java.util.function.Function;

import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.scene.control.ComboBox;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableCell;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.cell.CheckBoxTreeTableCell;
import javafx.scene.control.cell.ComboBoxTreeTableCell;
import javafx.scene.control.cell.TextFieldTreeTableCell;
import javafx.scene.input.Dragboard;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;
import javafx.util.Callback;
import javafx.util.converter.DefaultStringConverter;

/**
 * This UI provider is responsible to return the label, the icon, and the context menu for each cell
 * in the {@link DynamicTreeTableViewer} based on the data object and the column.
 */
public abstract class DynamicTreeTableUIProviderBase<T> {
	/**
	 * Retrieves the label to be displayed in the given position.
	 * 
	 * Note: An empty cell can only be editable, if an empty String is returned here instead of
	 * 'null'.
	 * 
	 * @param element
	 *            the element to be displayed in the current row
	 * @param column
	 *            the current column
	 * @return the label to be displayed in the given column
	 */
	public String getLabel(T element, int column) {
		return "";
	}

	/**
	 * @param element
	 *            the element to be displayed in the current row
	 * @param column
	 *            the current column
	 * @return the node to be displayed as the icon in the given column
	 */
	public Node getIconNode(T element, int column) {
		return null;
	}

	/**
	 * Returns the style that shall be applied to the {@link TreeItem} of the given element. Return
	 * {@code null} for the default style.
	 * 
	 * @param element
	 *            to apply a style to.
	 * @param column
	 *            column of to apply the style (the element is a row in the table).
	 * @return style of the element in the CSS format.
	 */
	public String getCellStyle(T element, int column) {
		if(element == null) {
			return null;
		}

		Color bgColor = getBackgroundColor(column, element);
		if(bgColor != null) {
			// Some transparency is necessary here, for correctly displaying the selection
			// bar when selecting a row. WIth out transparency, it would not be visible on
			// this cell.
			double alpha = 0.5;
			String colorStr = colorToRgbaString(bgColor, alpha);
			return "-fx-background-color: " + colorStr + ";";
		}
		return null;
	}

	/**
	 * Determines the background color for a given element in the specified column.
	 * 
	 * @param element
	 *            The element to specify the background color for.
	 * @param column
	 *            The column number to specify the background color for.
	 * 
	 * @return The {@link Color} to be used for the cell background.
	 */
	public Color getBackgroundColor(int column, T element) {
		return null;
	}

	/**
	 * @param element
	 *            the element to be displayed in the current row
	 * @param column
	 *            the current column
	 * @return the context menu in the given column
	 */
	public ContextMenu createContextMenu(T element, int column) {
		return null;
	}

	/**
	 * @param item
	 *            the tree element currently targeted by the drag location
	 * @param dragboard
	 *            the current {@link Dragboard}
	 * @return whether the drag-over event succeeds (i.e. should be consumed).
	 */
	public boolean dragOver(T item, Dragboard dragboard) {
		return false;
	}

	/**
	 * @param element
	 *            the element the content is dropped on
	 * @param dragboard
	 *            the current {@link Dragboard}
	 * @return whether the drop event completed (i.e. should be consumed).
	 */
	public boolean dropClipboardContent(T element, Dragboard dragboard) {
		return false;
	}

	/**
	 * Returns whether the given column is editable.
	 * 
	 * @param column
	 *            the column index
	 * @return whether the column is editable
	 */
	public boolean isEditable(int column) {
		return false;
	}

	/**
	 * Returns whether the given element is editable in the given column.
	 * 
	 * @param column
	 *            The column index
	 * @param element
	 *            The element to be checked
	 * @return Whether the column is editable
	 */
	public boolean isEditable(int column, T element) {
		return isEditable(column);
	}

	/**
	 * Updates the value after the given column was edited for the given element.
	 * 
	 * @param element
	 *            the element to be edited in the current row
	 * @param column
	 *            the currently edited column
	 * @param value
	 *            the edited value
	 */
	public void updateValue(T element, int column, Object value) {
		// ignored, since not editable
	}

	/**
	 * Notifies the provider about the selection of the given new value (and the deselection of the
	 * given old value.
	 * 
	 * @param oldValue
	 *            unselected value
	 * @param newValue
	 *            the selected value
	 */
	public void select(T oldValue, T newValue) {
		// ignored
	}

	/** Applies the editing support to the given column. */
	/* package */ final void applyToColumn(int i, TreeTableColumn<T, String> column) {
		if(!isEditable(i)) {
			column.setCellFactory(createReadOnlyCellFactory(i));
			column.setEditable(false);
			column.setOnEditCommit(null);
			return;
		}
		column.setCellFactory(createEditableCellFactory(i));
		column.setEditable(true);
		column.setOnEditStart(event -> {
			event.getEventType();
		});
		column.setOnEditCommit(event -> {
			T element = event.getRowValue().getValue();

			// The column number needs to be retrieved from the viewer, as the the
			// event.getTablePosition() does not take into account invisible columns for numbering.
			TreeTableColumn<T, String> tableColumn = event.getTableColumn();
			int colIndex = tableColumn.getTreeTableView().getColumns().indexOf(tableColumn);

			String value = event.getNewValue();
			updateValue(element, colIndex, value);
			column.getTreeTableView().refresh();
		});
	}

	/** Applies a checkbox editing support to the given checkbox column. */
	/* package */ final void applyToCheckboxColumn(int i, TreeTableColumn<T, Boolean> column) {
		column.setOnEditCommit(null);
		column.setEditable(isEditable(i));
		if(!isEditable(i)) {
			return;
		}

		// setOnEditCommit() does not work for CheckBoxTreeTableCell. This is why a custom
		// CellValueFactory has to be set here
		column.setCellValueFactory(new Callback<TreeTableColumn.CellDataFeatures<T, Boolean>, //
				ObservableValue<Boolean>>() {

			/** {@inheritDoc} */
			@Override
			public ObservableValue<Boolean>
					call(TreeTableColumn.CellDataFeatures<T, Boolean> param) {
				TreeItem<T> treeItem = param.getValue();
				T value = treeItem.getValue();
				String label = getLabel(value, i).trim().toLowerCase();
				// String to boolean transformations throw no error/exception that is why this has
				// to be captured manually here.
				if(!(label.equals("true") || label.equals("false"))) {
					throw new RuntimeException(
							"Wrong type in checkbox table cell. Expected boolean \"true\" or \"false\" but received \"" +
									label + "\".");
				}
				boolean b = Boolean.valueOf(label);
				SimpleBooleanProperty booleanProp = new SimpleBooleanProperty(b);

				// equal behavior as setOnEditCommit()
				booleanProp.addListener(new ChangeListener<Boolean>() {
					/** {@inheritDoc} */
					@Override
					public void changed(ObservableValue<? extends Boolean> observable,
							Boolean oldValue, Boolean newValue) {
						updateValue(value, i, newValue);
						column.getTreeTableView().refresh();
					}
				});
				return booleanProp;
			}
		});
		column.setCellFactory(createEditableCheckboxCellFactory(i));
	}

	/**
	 * Configures the given {@link TreeTableColumn} according to this UI provider.
	 * 
	 * @param columnIndex
	 *            The index number of the column to be configured.
	 * @param column
	 *            The {@link TreeTableColumn} to be configured.
	 * @param comboValueFactory
	 *            A {@link Function} that maps cell elements to a combo value map (data model
	 *            element -> label)
	 */
	/* package */ final void applyToComboColumn(int columnIndex, TreeTableColumn<T, String> column,
			Function<T, Map<?, String>> comboValueFactory) {
		column.setCellFactory(
				createEditableComboCellFactory(columnIndex, column, comboValueFactory));
	}

	/**
	 * Sets styling properties as background color for the given cell.
	 * 
	 * This method is not intended to be overwritten. Instead there are getters as
	 * {@link #getBackgroundColor(int, Object)}.
	 * 
	 * @param cell
	 *            The cell to be layouted.
	 * @param columnIndex
	 *            The number of the column in which this cell is contained in.
	 */
	private final void styleCell(TreeTableCell<T, ?> cell, int columnIndex) {
		if(cell.getTreeTableRow() != null) {
			T data = cell.getTreeTableRow().getItem();

			if(data != null) {
				final Node icon = getIconNode(data, columnIndex);
				cell.setGraphic(icon);
				cell.setStyle(getCellStyle(data, columnIndex));
			} else {
				// reset icon for cases, in which a row was styled before, but became empty by now.
				cell.setGraphic(null);
				cell.setStyle(null);
			}
		}
	}

	/** Converts a {@link Color} object into a rgba format as used in css. */
	private final static String colorToRgbaString(Color color, double alpha) {
		return String.format("rgba(%d,%d,%d,%s)", (int)(color.getRed() * 255),
				(int)(color.getGreen() * 255), (int)(color.getBlue() * 255), alpha);
	}

	/** Creates a cell factory for editable cells. */
	private Callback<TreeTableColumn<T, String>, TreeTableCell<T, String>>
			createEditableCellFactory(int colIndex) {
		return param -> {
			MyTextFieldTreeTableCell myTextFieldTreeTableCell =
					new MyTextFieldTreeTableCell(colIndex);
			styleCell(myTextFieldTreeTableCell, colIndex);
			return myTextFieldTreeTableCell;
		};
	}

	/** Creates a cell factory for editable cells. */
	private Callback<TreeTableColumn<T, Boolean>, TreeTableCell<T, Boolean>>
			createEditableCheckboxCellFactory(int colIndex) {
		return param -> {
			CheckBoxTreeTableCell<T, Boolean> checkBoxTreeTableCell =
					new CheckBoxTreeTableCell<T, Boolean>() {
						/** {@inheritDoc} */
						@Override
						public void updateItem(Boolean item, boolean empty) {
							super.updateItem(item, empty);
							styleCell(this, colIndex);
						}
					};
			return checkBoxTreeTableCell;
		};
	}

	/** Creates the cell factory for {@link ComboBox} editing. */
	private Callback<TreeTableColumn<T, String>, TreeTableCell<T, String>>
			createEditableComboCellFactory(int columnIndex, TreeTableColumn<T, String> column,
					Function<T, Map<?, String>> comboValueFactory) {
		// The list with the choices to be offered in the combo.
		ObservableList<String> items = observableArrayList();
		Callback<TreeTableColumn<T, String>, TreeTableCell<T, String>> comboCellFactory =
				param -> new ComboBoxTreeTableCell<T, String>(null, items) {

					/** {@inheritDoc} */
					@Override
					public void updateItem(String item, boolean empty) {
						super.updateItem(item, empty);

						styleCell(this, columnIndex);
					}

					/** {@inheritDoc} */
					@Override
					public void startEdit() {
						// null items are in cells, which cannot contain a value. Otherwise they
						// would contain an empty String.
						if(getItem() == null) {
							return;
						}

						super.startEdit();
					}
				};

		// Changes the choices of the combo for every editing of every cell dynamically.
		column.setOnEditStart(event -> {
			T rowEntry = event.getRowValue().getValue();
			Map<?, String> comboValueMap = comboValueFactory.apply(rowEntry);

			items.clear();
			items.addAll(comboValueMap.values());
		});

		column.setOnEditCommit(event -> {
			// The column number needs to be retrieved from the viewer, as the the
			// event.getTablePosition() does not take into account invisible columns for numbering.
			TreeTableColumn<T, String> tableColumn = event.getTableColumn();
			int colIndex = tableColumn.getTreeTableView().getColumns().indexOf(tableColumn);

			// Convert displayed value label into underlying data model's value
			String valueLabel = event.getNewValue();
			T rowEntry = event.getRowValue().getValue();
			Map<?, String> comboValueMap = comboValueFactory.apply(rowEntry);
			Map<String, ?> valueMapInversed = comboValueMap.entrySet().stream()
					.collect(toMap(Map.Entry::getValue, Map.Entry::getKey));

			updateValue(rowEntry, colIndex, valueMapInversed.get(valueLabel));
			column.getTreeTableView().refresh();
		});

		return comboCellFactory;
	}

	/** Custom table cell implementation to allow on key released validation. */
	private class MyTextFieldTreeTableCell extends TextFieldTreeTableCell<T, String> {
		/** The column index of this cell. */
		private final int columnIndex;

		/** Constructor. */
		public MyTextFieldTreeTableCell(int colIndex) {
			this.columnIndex = colIndex;
			setConverter(new DefaultStringConverter());
		}

		/** {@inheritDoc} */
		@Override
		public void startEdit() {
			if(getItem() == null || !isEditable() || !getTreeTableView().isEditable() ||
					!getTableColumn().isEditable()) {
				return;
			}
			super.startEdit();
			// the following code uses implementation details of
			// CellUtils.startEdit(...) and CellUtils.createTextfield(...)
			TextField tf = (TextField)getGraphic(); // CellUtils:228
			tf.setOnKeyReleased(t -> {
				DynamicTreeTableUIProviderBase<T>.MyTextFieldTreeTableCell mythis =
						MyTextFieldTreeTableCell.this;
				if(t.getCode() == KeyCode.ESCAPE) {
					mythis.cancelEdit();
					t.consume();
				} else {
					T item = mythis.getTreeTableRow().getItem();
					validateOnKeyReleased(t, columnIndex, item, tf.getText());
				}
			});
			tf.focusedProperty().addListener(new ChangeListener<Boolean>() {
				// If the focus is lost (user clicking elsewhere), the currently entered value is
				// committed.
				@Override
				public void changed(ObservableValue<? extends Boolean> val, Boolean oldVal,
						Boolean newVal) {
					if(!newVal) {
						commitEdit(tf.getText());
					}
				}
			});
		}

		/** {@inheritDoc} */
		@Override
		public void updateItem(String item, boolean empty) {
			super.updateItem(item, empty);
			styleCell(this, this.columnIndex);
		}

		/** {@inheritDoc} */
		@Override
		public void cancelEdit() {
			super.cancelEdit();
			T item = getTreeTableRow().getItem();
			validateOnCancelEdit(columnIndex, item);
		}
	}

	/**
	 * Called when editing a cell was canceled.
	 * 
	 * @param colIndex
	 *            the index of the column
	 * @param item
	 *            the row item
	 */
	public void validateOnCancelEdit(int colIndex, T item) {
		// nothing to do
	}

	/**
	 * Called when a key was released during editing. Implementors should consume the event.
	 * 
	 * @param event
	 *            the key released event
	 * @param colIndex
	 *            the index of the column
	 * @param item
	 *            the row item
	 * @param text
	 *            the current textfield content
	 */
	public void validateOnKeyReleased(KeyEvent event, int colIndex, T item, String text) {
		// nothing to do
	}

	/** Creates a cell factory for read-only cells. */
	private Callback<TreeTableColumn<T, String>, TreeTableCell<T, String>>
			createReadOnlyCellFactory(int colIndex) {
		return param -> {
			TreeTableCell<T, String> cell = new TreeTableCell<T, String>() {
				@Override
				protected void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);

					ContextMenu menu = null;
					if(!empty && item != null) {
						T data = this.getTreeTableRow().getItem();

						menu = createContextMenu(data, colIndex);
						if(IS_OS_LINUX && menu != null) {
							// JFX-builtin handling is non-operation in Linux.
							addContextMenuHandler(menu, data);
						}
					}

					this.setContextMenu(menu);
					styleCell(this, colIndex);
				}

				/**
				 * Adds a handler for showing the context menu. Duplicate of the same method in
				 * {@link DynamicTreeViewer} due to the absence of proper selection interfaces in
				 * JFX.
				 */
				private void addContextMenuHandler(ContextMenu menu, T element) {
					getTreeTableView().addEventHandler(MouseEvent.MOUSE_RELEASED, e -> {
						if(e.getButton() == MouseButton.SECONDARY) {
							TreeItem<T> selected =
									getTreeTableView().getSelectionModel().getSelectedItem();

							// We must check the element since all rows are evaluated.
							if(selected != null && selected.getValue() == element) {
								menu.show(this, e.getScreenX(), e.getScreenY());
								e.consume();
							}
						} else {
							// any other click cause hiding menu
							menu.hide();
						}
					});
				}
			};
			cell.textProperty().bind(cell.itemProperty());

			return cell;
		};
	}

	/**
	 * Applies this UIProvider to configure the given {@link TreeTableColumn}.
	 * 
	 * @param columnIndex
	 *            The index number of the column.
	 * @param column
	 *            The column to be configured.
	 */
	public void applyToTextColumn(int columnIndex, TreeTableColumn<T, String> column) {
		column.setCellFactory(ttColumn -> {
			DynamicTextFieldTreeTableCell<T> cell = new DynamicTextFieldTreeTableCell<T>() {

				/** {@inheritDoc} */
				@Override
				protected boolean canEdit(T element) {
					return DynamicTreeTableUIProviderBase.this.isEditable(columnIndex, element);
				}

				/** {@inheritDoc} */
				@Override
				protected void setValueOnLeavingCell(T element, String currentValue) {
					updateValue(element, columnIndex, currentValue);
				}

				/** {@inheritDoc} */
				@Override
				public void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);
					styleCell(this, columnIndex);
				}
			};

			styleCell(cell, columnIndex);

			return cell;
		});
	}
}
